//
//  PhobosDataManager.m
//  GMCache
//
//  Created by Locus on 2020/1/21.
//

#import "PhobosDataManager.h"
#import "PhobosSendManager.h"
#import "PhobosConfig.h"
#import "PhobosUtil.h"
#import <MJExtension/MJExtension.h>
#import <mach/mach_time.h>

@implementation PhobosDataManager

static NSManagedObjectContext *PhobosDefaultContext;
static NSUInteger PhobosToBeSendEntitiesCount = 0;

+ (void)initialize {
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        [self setupSDK];
        [self resetData];
    });
}

+ (NSUInteger)fetchCountOfToBeSendEntities {
    return PhobosToBeSendEntitiesCount;
}

+ (NSArray<PhobosSendDataEntity *> *)fetchToBeSendDataEntities {
    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"status = %d or status = %d", PhobosDataSendStatusToBeSend, PhobosDataSendStatusError];
    return [self fetchDataEntitiesWithPredicate:predicate];
}

+ (NSArray<PhobosSendDataEntity *> *)fetchDataEntitiesWithPredicate:(NSPredicate *)predicate {
    __block NSArray *results = nil;
    [[self PhobosContext] performBlockAndWait:^{
        results = [self noWaitFetchDataEntitiesWithPredicate:predicate];
    }];
    return results;
}

+ (NSArray<PhobosSendDataEntity *> *)fetchToBeSendDataEntitiesAndUpdateWithSendStatus:(PhobosDataSendStatus)sendStatus {
    __block NSArray *results = nil;
    [[self PhobosContext] performBlockAndWait:^{
        NSPredicate *predicate = [NSPredicate predicateWithFormat:@"status = %d or status = %d", PhobosDataSendStatusToBeSend, PhobosDataSendStatusError];
        results = [self noWaitFetchDataEntitiesWithPredicate:predicate];
        [self noWaitUpdateEntities:results sendStatus:sendStatus completion:nil];
    }];
    return results;
}

+ (void)insertData:(NSDictionary *)data sendAPI:(NSString *)sendAPI completion:(MRSaveCompletionHandler)completion {
    if (!sendAPI || [sendAPI isEqualToString:@""] || !data) {
        return;
    }
    [[self PhobosContext] performBlockAndWait:^{
        PhobosSendDataEntity *entity = [PhobosSendDataEntity MR_createEntityInContext:[self PhobosContext]];
        entity.data = [data mj_JSONData];
        entity.api = sendAPI;
        entity.status = PhobosDataSendStatusToBeSend;
        entity.id = mach_absolute_time();
        PhobosToBeSendEntitiesCount++;
         [self saveWithCompletion:completion];
    }];
}

+ (void)updateDataEntities:(NSArray<PhobosSendDataEntity *> *)entities sendStatus:(PhobosDataSendStatus)sendStatus completion:(MRSaveCompletionHandler)completion {
    if (entities.count > 0) {
        [[self PhobosContext] performBlockAndWait :^{
            [self noWaitUpdateEntities:entities sendStatus:sendStatus completion:completion];
        }];
    }
}

+ (void)deleteAllEntities {
    [[self PhobosContext] performBlockAndWait:^{
        PhobosToBeSendEntitiesCount = 0;
        [PhobosSendDataEntity MR_truncateAllInContext:[self PhobosContext]];
        [self saveWithCompletion:nil];
    }];
}

#pragma mark - private
/**
 * 初始化sdk
 */
+ (void)setupSDK {
    // 创建 NSManagedObjectContext，供埋点库访问CoreData库使用
    PhobosDefaultContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSPrivateQueueConcurrencyType];
    NSBundle *bundle = [NSBundle bundleForClass:[self class]];
    // 获取埋点数据的实体
    NSURL *modelURL = [bundle URLForResource:@"GMPhobosData" withExtension:@"momd"];
    NSManagedObjectModel *model = [[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL];
    
    // 创建持久化存储调度器
    NSPersistentStoreCoordinator *coordinator = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:model];
    
    // 创建并关联SQLite数据库文件，如果已经存在则不会重复创建
    NSString *dataPath = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES).lastObject;
    dataPath = [dataPath stringByAppendingFormat:@"/%@.sqlite",@"GMPhobos"];
    
    [coordinator addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:[NSURL fileURLWithPath:dataPath] options:nil error:nil];
    
    // 设置storeCoordinator
    [self PhobosContext].persistentStoreCoordinator = coordinator;
}

/*
 * 重置数据库中的数据和待发送数量
 */
+ (void)resetData {
    [[self PhobosContext] performBlockAndWait:^{
        /** 将上次没有获取到发送结果的数据的状态修改为发送失败，待下次重新发送 */
        NSPredicate *predicate = [NSPredicate predicateWithFormat:@"status = %d", PhobosDataSendStatusSending];
        NSArray<PhobosSendDataEntity *> *entities = [self noWaitFetchDataEntitiesWithPredicate:predicate];
        [self noWaitUpdateEntities:entities sendStatus:PhobosDataSendStatusError completion:nil];
        
        /** 将发送成功的数据删除 */
        NSPredicate *finishPredicate = [NSPredicate predicateWithFormat:@"status = %d", PhobosDataSendStatusFinish];
        [PhobosSendDataEntity MR_deleteAllMatchingPredicate:finishPredicate inContext:[self PhobosContext]];
        [self saveWithCompletion:nil];
        
        /** 设置未发送数据数量 */
        NSPredicate *countPredicate = [NSPredicate predicateWithFormat:@"status = %d or status = %d", PhobosDataSendStatusToBeSend, PhobosDataSendStatusError];
        PhobosToBeSendEntitiesCount = [[PhobosSendDataEntity MR_numberOfEntitiesWithPredicate:countPredicate inContext:[self PhobosContext]] integerValue];
    }];
}

/**
 * 在批量修改和删除时维护PhobosToBeSendEntitiesCount字段
 */
+ (void)updatePhobosCountOfToBeSendEntitiesWithStatus:(PhobosDataSendStatus)sendStatus changeCount:(NSUInteger)changeCount {
    if (sendStatus == PhobosDataSendStatusSending) {
        PhobosToBeSendEntitiesCount -= changeCount;
    } else if (sendStatus == PhobosDataSendStatusError) {
        PhobosToBeSendEntitiesCount += changeCount;
    }
}

/**
* 不使用同步线程安全方案[NSManagedObjectContext performBlockAndWait]获取谓词相应的数据
* 不对外使用，使用该方法时，需要保障线程同步
*/
+ (NSArray<PhobosSendDataEntity *> *)noWaitFetchDataEntitiesWithPredicate:(NSPredicate *)predicate {
    NSFetchRequest *request = [PhobosSendDataEntity MR_createFetchRequestInContext:[self PhobosContext]];
    [request setPredicate:predicate];
    NSError *error = nil;
    return [[self PhobosContext] executeFetchRequest:request error:&error];
}

/**
 * 不使用同步线程安全方案[NSManagedObjectContext performBlockAndWait]去修改实体数据sendStatus
 * 不对外使用，使用该方法时，需要保障线程同步
 */
+ (void)noWaitUpdateEntities:(NSArray<PhobosSendDataEntity *> *)entities sendStatus:(PhobosDataSendStatus)sendStatus completion:(MRSaveCompletionHandler)completion {
    [self updatePhobosCountOfToBeSendEntitiesWithStatus:sendStatus changeCount:entities.count];
    [entities enumerateObjectsUsingBlock:^(PhobosSendDataEntity *obj, NSUInteger idx, BOOL * _Nonnull stop) {
        obj.status = sendStatus;
    }];
    [self saveWithCompletion:completion];
}

+ (NSManagedObjectContext *)PhobosContext {
    return PhobosDefaultContext ?: [NSManagedObjectContext MR_defaultContext];
}

/**
 * 在保存完成后调用的完成块。如果发生错误，块将以“BOOL”和“NSError”实例的形式传递成功状态。总是在主队列上调用。
 */
+ (void)saveWithCompletion:(MRSaveCompletionHandler)completion {
    [[self PhobosContext] MR_saveOnlySelfWithCompletion:completion];
}

@end
